#   ipmkchains.py -- Linux firewall rule manipulation library
#   Copyright (C) 2000  Bruce Guenter <bruceg@em.ca>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import random
import re
import string
import socket
import time
import types

# Built in chain names
INPUT = intern('input')
FORWARD = intern('forward')
OUTPUT = intern('output')

CHAINS = [INPUT, FORWARD, OUTPUT]

# Predefined target names
ACCEPT   = intern('ACCEPT')
REJECT   = intern('REJECT')
DENY     = intern('DENY')
MASQ     = intern('MASQ')
REDIRECT = intern('REDIRECT')
RETURN   = intern('RETURN')

# Predefined protocol types
ICMP=1
TCP=6
UDP=17

icmp_types = {
    'echo-reply':(0,-1),
    'pong':(0,-1),
    'destination-unreachable':(3,-1),
    'network-unreachable':(3,0),
    'host-unreachable':(3,1),
    'protocol-unreachable':(3,2),
    'port-unreachable':(3,3),
    'fragmentation-needed':(3,4),
    'source-route-failed':(3,5),
    'network-unknown':(3,6),
    'host-unknown':(3,7),
    'network-prohibited':(3,9),
    'host-prohibited':(3,10),
    'TOS-network-unreachable':(3,11),
    'TOS-host-unreachable':(3,12),
    'communication-prohibited':(3,13),
    'host-precedence-violation':(3,14),
    'precedence-cutoff':(3,15),
    'source-quench':(4,-1),
    'redirect':(5,-1),
    'network-redirect':(5,0),
    'host-redirect':(5,1),
    'TOS-network-redirect':(5,2),
    'TOS-host-redirect':(5,3),
    'echo-request':(8,-1),
    'ping':(8,-1),
    'router-advertisement':(9,-1),
    'router-solicitation':(10,-1),
    'time-exceeded':(11,-1),
    'ttl-exceeded':(11,-1),
    'ttl-zero-during-transit':(11,0),
    'ttl-zero-during-reassembly':(11,1),
    'parameter-problem':(12,-1),
    'ip-header-bad':(12,0),
    'required-option-missing':(12,1),
    'timestamp-request':(13,-1),
    'timestamp-reply':(14,-1),
    'address-mask-request':(17,-1),
    'address-mask-reply':(18,-1),
    }

###############################################################################

_RX_DOTQ = re.compile(r'^(\d+)\.(\d+)\.(\d+)\.(\d+)$')

def hexint(str):
    return string.atoi(str, 16)

def hex2ip(str):
    return (string.atoi(str[0:2], 16),
            string.atoi(str[2:4], 16),
            string.atoi(str[4:6], 16),
            string.atoi(str[6:8], 16))

def ip2hex(ip):
    return '%02x%02x%02x%02x' % ip

def mask2ip(int):
    bits = 0xffffffff << (32-int)
    return ((bits >> 24) & 0xff,
            (bits >> 16) & 0xff,
            (bits >> 8) & 0xff,
            bits & 0xff)

def auto2ip(data):
    td = type(data)
    if td == types.TupleType:
        return data
    if td == types.ListType:
        return tuple(data)
    if td == types.StringType:
        return tuple(map(int, _RX_DOTQ.match(data).groups()))
    if td == types.IntType:
        return mask2ip(data)

def ip2str(ip):
    return '%d.%d.%d.%d' % ip

###############################################################################

FLAG1_LOG = 0x01
FLAG1_SYN = 0x02
FLAG1_FRAG = 0x04
FLAG1_SETMARK = 0x08
FLAG1_ANYIFACE = 0x10
FLAG1_OUTPUT = 0x20

FLAG2_SRCIP = 0x01
FLAG2_DSTIP = 0x02
FLAG2_PROTO = 0x04
FLAG2_SRCPORT = 0x08
FLAG2_DSTPORT = 0x10
FLAG2_IFACE = 0x20
FLAG2_SYN = 0x40
FLAG2_FRAG = 0x80

_RX_ADDR = re.compile(r'^(.+)/(.+)->(.+)/(.+)$')
_RX_PORTS = re.compile(r'^(\d+)-(\d+)$')
_RX_WHITESPACE = re.compile(r'\s+')

class Rule:
    def __init__(self, **values):
        defaults = {
            'target': '-',
            'proto': 0,
            'flags1': 0x08,
            'flags2': 0x00,
            'tosx': 0x00,
            'tosa': 0xff,
            'iface': '-',
            'mark': 0,
            'maxsize': 0,
            'srcip': (0,0,0,0),
            'srcmask': (0,0,0,0),
            'srcport1': 0,
            'srcport2': 65535,
            'dstip': (0,0,0,0),
            'dstmask': (0,0,0,0),
            'dstport1': 0,
            'dstport2': 65535,
            'redirect': 0,
            }
        for (key,value) in defaults.items():
            setattr(self, key, value)
        for (key,value) in values.items():
            if not defaults.has_key(key):
                raise AttributeError, "'%s' is an invalid value name"%key
            setattr(self, key, value)
        self.srcip = auto2ip(self.srcip)
        self.srcmask = auto2ip(self.srcmask)
        self.dstip = auto2ip(self.dstip)
        self.dstmask = auto2ip(self.dstmask)
        
    def __cmp__(self, that):
        for (key,val) in self.__dict__.items():
            c = cmp(val, getattr(that, key))
            if c:
                return c
        return 0

    def cmdpart(self, name, negflag, value):
        if self.flags2 & negflag:
            return ' --%(name)s ! %(value)s' % vars()
        else:
            return ' --%(name)s %(value)s' % vars()

        
    def command(self):
        # NOTE: only output ports if protocol is non-zero
        # NOTE: handle flags
        s = '--jump %s' % self.target
        s = s + self.cmdpart('proto', FLAG2_PROTO, self.proto)
        s = s + ' --TOS %x %x' % (self.tosa, self.tosx)
        if not (self.flags1 & FLAG1_ANYIFACE):
            s = s + self.cmdpart('interface', FLAG2_IFACE, self.iface)
        if self.flags1 & FLAG1_SETMARK or self.mark < 0:
            s = s + ' --mark %d' % self.mark
        else:
            s = s + ' --mark +%d' % self.mark
        s = s + self.cmdpart('source', FLAG2_SRCIP,
                             '%d.%d.%d.%d/%d.%d.%d.%d'%(
            self.srcip+self.srcmask))
        if self.proto:
            s = s + self.cmdpart('source-port', FLAG2_SRCPORT,
                                 '%d:%d' % (self.srcport1, self.srcport2))
        s = s + self.cmdpart('destination', FLAG2_DSTIP,
                             '%d.%d.%d.%d/%d.%d.%d.%d'%(
            self.dstip+self.dstmask))
        if self.proto:
            s = s + self.cmdpart('destination-port', FLAG2_DSTPORT,
                                 '%d:%d' % (self.dstport1, self.dstport2))
        if self.flags1 & FLAG1_LOG: s = s + ' --log'
        if self.flags1 & FLAG1_SYN: s = s + ' --syn'
        if self.flags1 & FLAG1_FRAG: s = s + ' --fragment'
        if self.flags1 & FLAG1_OUTPUT:
            s = s + ' --output %d' % self.maxsize
        return s

    def raw(self, chain):
        return ("%s %s/%s->%s/%s %s %x %x %d 0 0 0 0 "
                " %d-%d %d-%d A%x X%x %08x %d %d %s" % (
            chain, ip2hex(self.srcip), ip2hex(self.srcmask),
            ip2hex(self.dstip), ip2hex(self.dstmask),
            self.iface, self.flags1, self.flags2, self.proto,
            self.srcport1, self.srcport2, self.dstport1, self.dstport2,
            self.tosa, self.tosx, self.redirect, self.mark,
            self.maxsize, self.target ) )

def raw2rule(line):
    line = string.strip(line)
    (chain,
     addrs,iface,flags1,flags2,proto,phi,plo,bhi,blo,
     srcports,dstports,tosa,tosx,
     redirect,mark,maxsize,target) = _RX_WHITESPACE.split(line)
    (srcip,srcmask,
     dstip,dstmask) = map(hex2ip, _RX_ADDR.match(addrs).groups())
    (srcport1,srcport2) = map(int, _RX_PORTS.match(srcports).groups())
    (dstport1,dstport2) = map(int, _RX_PORTS.match(dstports).groups())
    
    flags1 = hexint(flags1)
    flags2 = hexint(flags2)
    proto = int(proto)
    tosa = hexint(tosa[1:])
    tosx = hexint(tosx[1:])
    redirect = hexint(redirect)
    mark = int(mark)
    maxsize = int(maxsize)
    return (chain,
            Rule(target=target, proto=proto,
                 flags1=flags1, flags2=flags2,
                 tosa=tosa, tosx=tosx, iface=iface,
                 mark=mark, maxsize=maxsize,
                 srcip=srcip, srcmask=srcmask,
                 srcport1=srcport1, srcport2=srcport2,
                 dstip=dstip, dstmask=dstmask,
                 dstport1=dstport1, dstport2=dstport2,
                 redirect=redirect )
            )
###############################################################################

_RX_MASKEDIP = re.compile('^(\d+\.\d+\.\d+\.\d+)/(\d+)$')

class Address:
    def __init__(self, ip, mask=None, negate=None):
        # If the first arg is the same as this, use it to set the values of
        # IP, mask, and negate only if they aren't also set.
        if type(ip) is type(self):
            if mask is None:
                mask = ip.mask
            if negate is None:
                negate = ip.negate
            ip = ip.ip
        # Set defaults for network mask and negate flag
        if mask is None:
            mask = (255,255,255,255)
        if negate is None:
            negate = 0
        # Convert strings of the form "a.b.c.d/m" to IP/mask
        if type(ip) is types.StringType:
            m = _RX_MASKEDIP.match(ip)
            if m:
                ip = auto2ip(m.group(1))
                mask = int(m.group(2))
        self.ip = auto2ip(ip)
        self.mask = auto2ip(mask)
        self.negate = negate

# The following is included only for backwards compatibility
address = Address

###############################################################################

_commands = [ ]
_commands_post = [ ]
_commands_pre = [ ]

_dst_defaults = { }

_dst_chains = { }

def command_pre(cmd):
    global _commands_pre
    _commands_pre.append(cmd)

def command(cmd):
    global _commands
    _commands.append(cmd)

def command_post(cmd):
    global _commands_post
    _commands_post.append(cmd)

command_pre('PATH=/sbin:/usr/sbin:/usr/local/sbin:$PATH')
command_pre('export PATH')

def commands():
    global _commands_pre
    global _commands
    global _commands_post
    return _commands_pre + _commands + _commands_post

def _tmpfile(base):
    while 1:
        filename = '/tmp/%s.tmp.%d.%f.%f' % (base, os.getpid(), time.time(),
                                             random.uniform(0, 1000000000))
        try:
            fd = os.open(filename, os.O_CREAT | os.O_EXCL, 0600)
            os.close(fd)
            break
        except OSError:
            continue
    return filename

def _read_fwchains(fwchains = '/proc/net/ip_fwchains',
                   fwnames = '/proc/net/ip_fwnames'):
    defaults = { }
    entries = { }
    for line in open(fwnames, 'r').readlines():
        (chain,default,rest) = string.split(line, ' ', 2)
        defaults[chain] = default
        entries[chain] = [ ]
    for line in open(fwchains, 'r').readlines():
        (chain,entry) = raw2rule(line)
        try:
            entries[chain].append(entry)
        except:
            entries[chain] = [entry]
    return defaults, entries

def flush_changes(lineno, srclen, inserts, deletes):
    while len(inserts) > 0 and len(deletes) > 0:
        (chain,rule) = raw2rule(inserts[0])
        cmd = 'ipchains -R %s %d %s' % (chain, lineno, rule.command())
        command(cmd)
        lineno = lineno + 1
        del inserts[0]
        del deletes[0]
    while len(inserts) > 0:
        (chain,rule) = raw2rule(inserts[0])
        cmd = rule.command()
        if lineno > srclen:
            cmd = 'ipchains -A %s %s' % (chain, cmd)
        elif lineno == 1:
            cmd = 'ipchains -I %s %s' % (chain, cmd)
        else:
            cmd = 'ipchains -I %s %d %s' % (chain, lineno, cmd)
        command(cmd)
        srclen = srclen + 1
        lineno = lineno + 1
        del inserts[0]
    while len(deletes) > 0:
        (chain,rule) = raw2rule(deletes[0])
        cmd = 'ipchains -D %s %d' % (chain, lineno)
        command(cmd)
        srclen = srclen - 1
        del deletes[0]
    return lineno, srclen

def _save_chain(name, chain, filename):
    f = open(filename, 'w')
    for i in range(len(chain)):
        f.write(chain[i].raw(name))
        f.write('\n')
    f.close()

def _diff_chains(chain, src, dst):
    tmpfile_src = _tmpfile('ipmkchains')
    tmpfile_dst = _tmpfile('ipmkchains')

    _save_chain(chain, src, tmpfile_src)
    _save_chain(chain, dst, tmpfile_dst)

    p = os.popen('diff -u -b %s %s' % (tmpfile_src, tmpfile_dst), 'r')
    result = map(string.rstrip, p.readlines())
    p.close()
    os.unlink(tmpfile_src)
    os.unlink(tmpfile_dst)
    return result

def _scan_chain(chain, src, dst, verbose):
    srclen = len(src)

    lineno = 0
    inserts = []
    deletes = []
    for line in _diff_chains(chain, src, dst):
        if verbose:
            command('# ' + line)
        if line[:4] == '--- ' or line[:4] == '+++ ':
            continue
        m = re.match(r'^@@ -(\d+),(\d+) ', line)
        if m:
            (lineno,srclen) = flush_changes(lineno, srclen, inserts, deletes)
            lineno = int(m.group(1)) or 1
        elif line[0] == '-':
            deletes.append(line[1:])
        elif line[0] == '+':
            inserts.append(line[1:])
        elif line[0] == ' ':
            (lineno,srclen) = flush_changes(lineno, srclen, inserts, deletes)
            lineno = lineno + 1
    flush_changes(lineno, srclen, inserts, deletes)

def _set_default(chain, srcpolicy, dstpolicy):
    if srcpolicy <> dstpolicy:
        command('ipchains --policy %s %s' % (chain, dstpolicy))

def default(chain, target):
    if chain not in CHAINS:
        raise ValueError, "Can only set policy on built-in chains."
    global _dst_defaults
    _dst_defaults[chain] = target

def rule(chain, target, proto=None,
         src=None, srcport=None, dst=None, dstport=None,
         iface=None, mark=0, output=0, tos=None, fragment=None,
         redirect=0, syn=None, log=None, bidirectional=None,
         icmptype=None):
    if proto is None:
        proto = 0
    if src:
        src = Address(src)
    else:
        src = Address((0,0,0,0),(0,0,0,0),0)
    if dst:
        dst = Address(dst)
    else:
        dst = Address((0,0,0,0),(0,0,0,0),0)
    if not srcport:
        srcport1 = 0
        srcport2 = 65535
    else:
        srcport1 = srcport2 = srcport
    if not dstport:
        dstport1 = 0
        dstport2 = 65535
    else:
        dstport1 = dstport2 = dstport
    if not tos:
        tos = (0xff, 0x00)

    if icmptype:
        if proto <> ICMP:
            raise ValueError, "Protocol must by ICMP to use ICMP types."
        if icmp_types.has_key(icmptype):
            icmptype = icmp_types[icmptype]
        if type(icmptype) is not types.TupleType:
            raise ValueError, "Invalid ICMP type value '%s'" % icmptype
        if icmptype[0] >= 0:
            srcport1 = srcport2 = icmptype[0]
        else:
            srcport1 = 0
            srcport2 = 65535
        if icmptype[1] >= 0:
            dstport1 = dstport2 = icmptype[1]
        else:
            dstport1 = 0
            dstport2 = 65535

    flags2 = 0
    flags1 = 0
    if not iface:
        flags1 = flags1 | FLAG1_ANYIFACE
        iface = '-'
    if log:
        flags1 = flags1 | FLAG1_LOG
    if syn:
        flags1 = flags1 | FLAG1_SYN
        if syn < 0:
            flags2 = flags2 | FLAG2_SYN
    if fragment:
        flags1 = flags1 | FLAG1_FRAG
        if fragment < 0:
            flags2 = flags2 | FLAG2_FRAG
    if mark:
        flags1 = flags1 | FLAG1_SETMARK
    if output:
        flags1 = flags1 | FLAG1_OUTPUT
    if src.negate:
        flags2 = flags2 | FLAG2_SRCIP
    if dst.negate:
        flags2 = flags2 | FLAG2_DSTIP
    if srcport < 0:
        srcport = -srcport
        flags2 = flags2 | FLAG2_SRCPORT
    if dstport < 0:
        dstport = -dstport
        flags2 = flags2 | FLAG2_DSTPORT
    if iface[0] == '-' and len(iface) > 1:
        iface = iface[1:]
        flags2 = flags2 | FLAG2_IFACE

    r = Rule(target=target, proto=proto, flags1=flags1, flags2=flags2,
             tosa=tos[0], tosx=tos[1], iface=iface, mark=mark,
             maxsize=output, redirect=redirect,
             srcip=src.ip, srcmask=src.mask,
             srcport1=srcport1, srcport2=srcport2,
             dstip=dst.ip, dstmask=dst.mask,
             dstport1=dstport1, dstport2=dstport2)
    global _dst_chains
    try:
        _dst_chains[chain].append(r)
    except:
        _dst_chains[chain] = [r]
    if bidirectional:
        r = Rule(target=target, proto=proto, flags1=flags1, flags2=flags2,
                 tosa=tos[0], tosx=tos[1], iface=iface, mark=mark,
                 maxsize=output, redirect=redirect,
                 srcip=dst.ip, srcmask=dst.mask,
                 srcport1=dstport1, srcport2=dstport2,
                 dstip=src.ip, dstmask=src.mask,
                 dstport1=srcport1, dstport2=srcport2)
        _dst_chains[chain].append(r)

def install(test=0, verbose=0):
    src_defaults, src_chains = _read_fwchains()
    global _dst_chains
    global _dst_defaults
    for chain in src_chains.keys():
        if not _dst_chains.has_key(chain) and chain not in CHAINS:
            command_post('ipchains --flush %s' % chain)
            command_post('ipchains --delete-chain %s' % chain)
    for chain in _dst_chains.keys():
        if not src_chains.has_key(chain):
            src_chains[chain] = []
            command_pre('ipchains --new %s' % chain)
        if chain in CHAINS:
            # Set default before scanning chain
            _set_default(chain, src_defaults.get(chain, DENY),
                         _dst_defaults.get(chain, DENY))
        _scan_chain(chain, src_chains[chain], _dst_chains[chain], verbose)
    list = string.join(commands(), '\n')
    if test:
        print list
    else:
        os.system(list)

